// Animated "3D Game of Life" with random  Evgeny Demidov  6 Nov 2000
// I used "3D balls" from "MoleculeViewer" by James Gosling
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.StringTokenizer;
public class ALife extends java.applet.Applet implements MouseListener,
  MouseMotionListener, KeyListener, ActionListener, ItemListener, Runnable {
  Thread thCycle;
  int w,h,w2,h2,  mx0,my0, delay=1000, it,
    r1=2, r2=5 ,r3=3, r4=2,
    n=11, nvert, n1, n2,  N1, N2, zs[],
    R=10, Rl=255, Gl=0, Bl=0, R2=2*R, hx=R/4, hy=hx;
  float ax[],ay[],az[];
  double fiX=.3, fiY=.3, dfi, scaleX,scaleZ,Ro;
  Image buffImage, balls[];
  Graphics buffGraphics;
  private final static int bgGrey = 192;
  private final static int nBalls = 16;
  byte L[], Sum[][][]; // So[][][];
  boolean bCycle = false;
  Label     lbN, lbDelay, lbR;
  TextField tfN, tfDelay, tfR;
  Button   btRun, btRand;
  Checkbox  cbPeriodic;  boolean bPeriodic = false;

public void init() {
  w = getSize().width;   h = getSize().height;
  dfi = 6.28 / w;   w2 = w/2; h2 = h/2;
  buffImage = createImage(w, h);
  buffGraphics = buffImage.getGraphics();
  setBackground(Color.white);
  Setup();
  String s=getParameter("n");
  if (s != null) n = Integer.parseInt(s);
  setupN();
  s=getParameter("rules"); if (s != null){
    StringTokenizer st = new StringTokenizer(s);
    r1 = Integer.parseInt(st.nextToken());
    r2 = Integer.parseInt(st.nextToken());
    r3 = Integer.parseInt(st.nextToken());
    r4 = Integer.parseInt(st.nextToken());}
  lbN = new Label("N");             add(lbN);
  tfN = new TextField( "" + n, 2);  add(tfN);
  lbR = new Label("R");             add(lbR);
  tfR = new TextField( ""+r1+" "+r2+" "+r3+" "+r4, 8);  add(tfR);
  cbPeriodic = new Checkbox("Periodic");    add(cbPeriodic);
  lbDelay = new Label("Delay");             add(lbDelay);
  tfDelay = new TextField( "" + delay, 4);  add(tfDelay);
  tfN.addKeyListener(this);  tfR.addKeyListener(this);
  tfDelay.addKeyListener(this);
  btRun = new Button("Run ");   btRun.addActionListener(this);
  add(btRun);
  btRand = new Button("Rand"); btRand.addActionListener(this);
  add(btRand);
  addMouseListener(this);
  addMouseMotionListener(this);
  cbPeriodic.addItemListener(this);
}
void setupN(){
  nvert=n*n*n;
  n1=n+1; n2=n+2;  N1=n1*n1*n1; N2=n2*n2*n2;
  scaleX = w*.6/n; scaleZ = nBalls/(1.75*n);
  Ro = ((double)n/2 - .5);
  zs = new int[nvert];
  ax = new float[nvert];  ay = new float[nvert];  az = new float[nvert];
  for (int i = 0; i < nvert; i++) zs[i] = i;
  L = new byte[nvert];
  Sum = new byte[n2][n2][n2]; // So = new byte[n2][n2][n2];
  it = 0;
  int t=(1+n*(1+n))*(n/2);   L[t]=1; L[t-1]=L[t+1]=1;
  rotate();
}
public void destroy() {
  removeMouseListener(this);
  removeMouseMotionListener(this);
}
public void run() {
  while(true) {
    if ( Thread.currentThread() != thCycle ) return;
    cycle();
    repaint();
    try {  Thread.sleep( delay );
    } catch (InterruptedException e) {}
  }
}
public void stop() {
  thCycle = null;
}
public void mouseClicked(MouseEvent e){}       // event handling
public void mousePressed(MouseEvent e) {
  mx0 = e.getX();  my0 = e.getY();
  e.consume();
}
public void mouseReleased(MouseEvent e){}
public void mouseEntered(MouseEvent e) {}
public void mouseExited(MouseEvent e)  {}
public void mouseDragged(MouseEvent e) {
  int x1 = e.getX();  int y1 = e.getY();
  fiY += dfi*(x1 - mx0);   mx0 = x1;
  fiX += dfi*(y1 - my0);   my0 = y1;
  rotate();
  repaint();
  e.consume();
}
public void mouseMoved(MouseEvent e) {}
public void actionPerformed(ActionEvent e){
  if ( e.getActionCommand().equals("Rand") ){
     for (int i = 0; i < nvert; i++) L[i] = (byte)(1.1*Math.random());
     rotate();  it = 0;
     repaint();}
  else{
   if (bCycle) {
     bCycle = false;  btRun.setLabel("Run "); thCycle = null;}
   else { bCycle = true;  btRun.setLabel("Stop");
     thCycle = new Thread(this); thCycle.start();}
  }
}
public void keyTyped(KeyEvent e) {}
public void keyPressed(KeyEvent e) {}
public void keyReleased(KeyEvent e) {
  final int keyEnter = 10;
  if (e.getKeyCode() == keyEnter) {
    int newN = n;
    try{
      delay = Integer.parseInt( tfDelay.getText() );
      newN = Integer.parseInt( tfN.getText() );
      StringTokenizer st = new StringTokenizer(tfR.getText());
      r1 = Integer.parseInt(st.nextToken());
      r2 = Integer.parseInt(st.nextToken());
      r3 = Integer.parseInt(st.nextToken());
      r4 = Integer.parseInt(st.nextToken());
    }catch ( NumberFormatException ne) {}
    if (newN != n){
      n = newN;
      setupN();
      repaint();}
  }
  e.consume();
}
public void itemStateChanged(ItemEvent e){
  bPeriodic = cbPeriodic.getState();
}
public void paint(Graphics g) {
  buffGraphics.clearRect(0, 0, w, h);
  int xo = w2-R,  yo = h2-R;
  for (int i = nvert-1; i >= 0; i--){
    int j = zs[i];
    if (L[j]==0){
      int xi = w2+(int)ax[j], yi = h2+(int)ay[j];
      buffGraphics.drawLine(xi,yi,xi,yi);}
    else{
      buffGraphics.drawImage(
        balls[(int)az[j]], xo+(int)ax[j], yo+(int)ay[j], this);}
  }
  g.drawImage(buffImage, 0, 0, this);
  showStatus( "Iteration=" + it);
}
public void update(Graphics g)   {  paint(g);  }

void cycle(){
//  System.arraycopy(So, 0, Sum, 0, So.length);
  it++;
  for (int i = 0; i < n2; i++)
   for (int j = 0; j < n2; j++)
    for (int k = 0; k < n2; k++) Sum[i][j][k] = 0;
  int p = 0;
  for (int i = 1; i < n1; i++)
   for (int j = 1; j < n1; j++)
    for (int k = 1; k < n1; k++){
     if (L[p++]==1){
      for (int io = -1; io < 2; io++)
       for (int jo = -1; jo < 2; jo++)
        for (int ko = -1; ko < 2; ko++) ++Sum[i+io][j+jo][k+ko];
      --Sum[i][j][k];
     }
    }
  if (bPeriodic){
   for (int i = 0; i < n2; i++)
    for (int j = 0; j < n2; j++){
      Sum[i][j][1] += Sum[i][j][n1];
      Sum[i][j][n] += Sum[i][j][0];}
   for (int i = 0; i < n2; i++)
    for (int j = 0; j < n2; j++){
      Sum[i][1][j] += Sum[i][n1][j];
      Sum[i][n][j] += Sum[i][0][j];}
   for (int i = 0; i < n2; i++)
    for (int j = 0; j < n2; j++){
      Sum[1][i][j] += Sum[n1][i][j];
      Sum[n][i][j] += Sum[0][i][j];}
  }
  p = 0;
  for (int i = 1; i < n1; i++)
   for (int j = 1; j < n1; j++)
    for (int k = 1; k < n1; k++){              // rules
     if ( (L[p]==0)&&(Sum[i][j][k]>=r1)&&(Sum[i][j][k]<=r2) ) L[p]=1;
     else if ( (L[p]!=0)&&((Sum[i][j][k]>r3)||(Sum[i][j][k]<r4)) ) L[p]=0;
     p++;
    }
  repaint();
}

void rotate(){
  int p=0;
  double cx = Math.cos(-fiX), cy = Math.cos(-fiY),
         sx = Math.sin(-fiX), sy = Math.sin(-fiY),
         dix = scaleX*cy, diy = -scaleX*sx*sy, diz = -scaleZ*cx*sy,
         djy = scaleX*cx, djz = -scaleZ*sx,
         dkx = scaleX*sy, dky = scaleX*sx*cy, dkz = scaleZ*cx*cy;
  double ix=-Ro*(dix+dkx), iy=-Ro*(diy+djy+dky), iz=-Ro*(diz+djz+dkz)+nBalls/2;
  for (int i = 0; i < n; i++){
    double jy=0, jz=0;
    for (int j = 0; j < n; j++){
      double kx=0, ky=0, kz=0;
      for (int k = 0; k < n; k++){
        ax[p] = (float)(ix + kx);
        ay[p] = (float)(iy + jy + ky);
        az[p++] = (float)(iz + jz + kz);
        kx += dkx;  ky += dky;  kz += dkz;}
      jy += djy;  jz += djz;}
    ix += dix;  iy += diy;  iz += diz;
  }
  for (int i = nvert - 1; --i >= 0;) {
    boolean flipped = false;
    for (int j = 0; j <= i; j++) {
      int a = zs[j],  b = zs[j + 1];
      if (az[a] > az[b]) {
        zs[j + 1] = a;   zs[j] = b;
        flipped = true;
      }
    }
    if (!flipped) break;
  }
}
int blend(int fg, int bg, float fgfactor) {
  return (int) (bg + (fg - bg) * fgfactor);
}
void Setup() {
  byte[] data = new byte[R2*R2];
  int mr = 0;
  for (int Y = R2; --Y >= 0;) {
    int x0 = (int) (Math.sqrt(R*R - (Y - R)*(Y - R)) + 0.5);
    int p = Y*R2 + R - x0;
    for (int X = -x0; X < x0; X++) {
      int x = X + hx;
      int y = Y - R + hy;
      int r = (int) (Math.sqrt(x * x + y * y) + 0.5);
      if (r > mr) mr = r;
      data[p++] = r <= 0 ? 1 : (byte) r;
    }
  }
  balls = new Image[nBalls];
  byte red[] = new byte[256];
  red[0] = (byte) bgGrey;
  byte green[] = new byte[256];
  green[0] = (byte) bgGrey;
  byte blue[] = new byte[256];
  blue[0] = (byte) bgGrey;
  for (int r = 0; r < nBalls; r++) {
    float b = (float) (r+1) / nBalls;
    for (int i = mr; i >= 1; --i) {
      float d = (float) i / mr;
      red[i]   = (byte) blend(blend(Rl, 255, d), bgGrey, b);
      green[i] = (byte) blend(blend(Gl, 255, d), bgGrey, b);
      blue[i]  = (byte) blend(blend(Bl, 255, d), bgGrey, b);
    }
    IndexColorModel model = new IndexColorModel(8, mr+1, red, green, blue, 0);
    balls[ nBalls - r - 1 ] = createImage(
                new MemoryImageSource(R2, R2, model, data, 0, R2));
  }
}

}
